/*
 * winhelp.c --
 *
 *	Tcl extension module to access HTMLHelp API.
 *	Provides the "winhelp" tcl command.
 *
 *	(c) 2000 <Christian.Werner@t-online.de>
 *
 *	See the file "license.terms" for information on
 *	usage and redistribution of this file, and for a
 *	DISCLAIMER OF ALL WARRANTIES.
 */

#include <windows.h>
#include <htmlhelp.h>
#include <tk.h>

#if (TCL_MAJOR_VERSION < 8)
#error "need Tcl 8.0 or higher"
#endif
#if (TCL_MAJOR_VERSION == 8) && (TCL_MINOR_VERSION < 1)
#define TCL80 1
#endif

/*
 * Structure for deferred start of HTML help control.
 */

typedef struct {
    Tk_Window tkwin;		/* Tk window for which to open HTML help */
    char *arg1, *arg2;		/* Arguments for HtmlHelp() */
} HandlerInfo;

/*
 * Things for Tcl 8.1 and above, eg encoding, threads.
 */

#ifndef TCL80
static int unicode = -1;
TCL_DECLARE_MUTEX(winhelpMutex)
#endif

/*
 * DANGER!!!DANGER!!!DANGER!!!DANGER!!!DANGER!!!
 *
 * This is not necesseraly portable to newer Tk versions
 * Alternative would be to #include platform specific headers.
 *
 * DANGER!!!DANGER!!!DANGER!!!DANGER!!!DANGER!!!
 */

#ifdef USE_TK_STUBS
typedef struct {
    int type;
    HWND handle;
    Tk_Window tkwin;
} TkWinWindow;
#define Tk_GetHWND(w) (((TkWinWindow *) w)->handle)
#else
extern HWND	Tk_GetHWND _ANSI_ARGS_((Window w));
#endif

/*
 * Local procedures.
 */

static HWND	CallHtmlHelp _ANSI_ARGS_((HWND hwndCaller, char *fileName,
					  UINT uCommand, char *data));
static void	MapHandler _ANSI_ARGS_((ClientData clientData,
					XEvent *eventPtr));
static int	GetTkWin _ANSI_ARGS_((Tk_Window mainwin,
				      Tcl_Interp *interp, char *name,
				      Tk_Window *winPtr));
static int	WinHelpCmd _ANSI_ARGS_((ClientData clientData,
					Tcl_Interp *interp, int argc,
                                        char **argv));

/*
 *----------------------------------------------------------------------
 *
 *  CallHtmlHelp --
 *
 *	This procedure is called to load the HTML Help DLL and call
 *	the HtmlHelp function therein.
 *
 *----------------------------------------------------------------------
 */

static HWND
CallHtmlHelp(hwndCaller, fileName, uCommand, data)
    HWND hwndCaller;
    char *fileName;
    UINT uCommand;
    char *data;
{
    static HINSTANCE inst = NULL;
    HWND result = NULL;
    FARPROC proc;

#ifndef TCL80
    Tcl_MutexLock(&winhelpMutex);
#endif
    if (inst == NULL) {
	inst = LoadLibraryA("hhctrl.ocx");
    }
#ifndef TCL80
    Tcl_MutexUnlock(&winhelpMutex);
#endif
    if (inst == NULL) {
	MessageBeep(MB_ICONEXCLAMATION);
	return result;
    }
#ifdef TCL80
    proc = GetProcAddress(inst, "HtmlHelpA");
    if (proc != NULL) {
	result = (HWND) proc(hwndCaller, fileName, uCommand, data);
    }
#else
    proc = GetProcAddress(inst, unicode ? "HtmlHelpW" : "HtmlHelpA");
    if (proc != NULL) {
	Tcl_DString fileDS, dataDS;
	char *dataPtr = NULL;
	Tcl_Encoding enc = unicode ? Tcl_GetEncoding(NULL, "unicode") : NULL;

	Tcl_DStringInit(&fileDS);
	Tcl_UtfToExternalDString(enc, fileName, -1, &fileDS);
	Tcl_DStringInit(&dataDS);
	if (data != NULL) {
	    Tcl_UtfToExternalDString(enc, data, -1, &dataDS);
	    dataPtr = Tcl_DStringValue(&dataDS);
	}
	result = (HWND) proc(hwndCaller, Tcl_DStringValue(&fileDS),
			     uCommand, dataPtr);
	Tcl_DStringFree(&fileDS);
	Tcl_DStringFree(&dataDS);
    }
#endif
    else {
	MessageBeep(MB_ICONEXCLAMATION);
    }
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 *  MapHandler --
 *
 *	This procedure is called when a toplevel is initially mapped.
 *      It starts up the HTML help control with the arguments taken
 *	from the HandlerInfo.
 *
 *----------------------------------------------------------------------
 */

static void
MapHandler(clientData, eventPtr)
    ClientData clientData;
    XEvent *eventPtr;
{
    if (eventPtr->type == MapNotify) {
	HandlerInfo *hiPtr = (HandlerInfo *) clientData;
	HWND hWnd;

	hWnd = (HWND) Tk_GetHWND(Tk_WindowId(hiPtr->tkwin));
	if (hWnd != NULL &&
	    (hWnd = GetParent(hWnd)) != NULL) {
	    CallHtmlHelp(hWnd, hiPtr->arg1, HH_DISPLAY_TOPIC, hiPtr->arg2);
	    Tk_DeleteEventHandler(hiPtr->tkwin, StructureNotifyMask,
				  MapHandler, clientData);
	    Tcl_Free((char *) hiPtr);
	}
    }
}

/*
 *----------------------------------------------------------------------
 *
 *  GetTkWin --
 *
 *	Given window's pathname this procedure returns the correspoding
 *	Tk_Window token for that window. It complains if the window is
 *	not a toplevel. In that case an error message is left in the
 *	interpreter.
 *
 *----------------------------------------------------------------------
 */

static int
GetTkWin(mainwin, interp, name, winPtr)
    Tk_Window mainwin;
    Tcl_Interp *interp;
    char *name;
    Tk_Window *winPtr;
{
    Tk_Window tkwin;

    tkwin = Tk_NameToWindow(interp, name, mainwin);
    if (tkwin == NULL) {
	return TCL_ERROR;
    }
    if (!Tk_IsTopLevel(tkwin)) {
	Tcl_AppendResult(interp, "\"", name, "\" is not a toplevel window",
			 (char *) NULL);
	return TCL_ERROR;
    }
    Tk_MakeWindowExist(tkwin);
    *winPtr = tkwin;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 *  WinHelpCmd --
 *
 *	This procedure implements the "winhelp" Tcl command.
 *
 *----------------------------------------------------------------------
 */

static int
WinHelpCmd(clientData, interp, argc, argv)
    ClientData clientData;
    Tcl_Interp *interp;
    int argc;
    char **argv;
{
    Tk_Window tkwin, mainwin = (Tk_Window) clientData;
    HWND hWnd;
    HandlerInfo *hiPtr;

    if (argc < 3 || argc > 4) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " pathName file ?file?\"", (char *) NULL);
	return TCL_ERROR;
    }
    if (GetTkWin(mainwin, interp, argv[1], &tkwin) != TCL_OK) {
	return TCL_ERROR;
    }
    if (Tk_IsMapped(tkwin)) {
	hWnd = (HWND) Tk_GetHWND(Tk_WindowId(tkwin));
	if (hWnd != NULL &&
	    (hWnd = GetParent(hWnd)) != NULL) {
	    CallHtmlHelp(hWnd, argv[2], HH_DISPLAY_TOPIC,
			 argc > 3 ? argv[3] : NULL);
	    return TCL_OK;
	}
    }

    /*
     * Window might not exist at this point. Start of the HTML help
     * control is deferred until window is mapped.
     */
    hiPtr = (HandlerInfo *) Tcl_Alloc(sizeof (HandlerInfo));
    hiPtr->tkwin = tkwin;
    hiPtr->arg1 = argv[2];
    hiPtr->arg2 = argc > 3 ? argv[3] : NULL;
    Tk_CreateEventHandler(tkwin, StructureNotifyMask, MapHandler,
			  (ClientData) hiPtr);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 *  Winhelp_Init --
 *
 *	Initialization procedure for this module; creates the "winhelp"
 *	Tcl command.
 *
 *----------------------------------------------------------------------
 */

int __declspec(dllexport)
Winhelp_Init(interp)
    Tcl_Interp *interp;
{
#ifndef TCL80
    OSVERSIONINFO os;

    Tcl_MutexUnlock(&winhelpMutex);
    if (unicode < 0) {
	os.dwOSVersionInfoSize = sizeof (os);
	GetVersionEx(&os);
	unicode = os.dwPlatformId == VER_PLATFORM_WIN32_NT;
    }
    Tcl_MutexUnlock(&winhelpMutex);
#endif
#ifdef USE_TCL_STUBS
    if (!Tcl_InitStubs(interp, "8.0", 0)) {
	return TCL_ERROR;
    }
#else
    if (Tcl_PkgRequire(interp, "Tcl", TCL_VERSION, 0) == NULL) {
	return TCL_ERROR;
    }
#endif
#ifdef USE_TK_STUBS
    if (!Tk_InitStubs(interp, "8.0", 0)) {
	return TCL_ERROR;
    }
#else
    if (Tcl_PkgRequire(interp, "Tk", TK_VERSION, 0) == NULL) {
	return TCL_ERROR;
    }
#endif
    Tcl_CreateCommand(interp, "winhelp", WinHelpCmd,
		      (ClientData) Tk_MainWindow(interp),
		      (Tcl_CmdDeleteProc *) NULL);
    return Tcl_PkgProvide(interp, "winhelp", "1.0");
}

